/****************************************************************
	LLPAM Emulator
	(built-in predicates)
	by Naoyuki Tamura (tamura@kobe-u.ac.jp)
	   Dept. CS, Kobe University
****************************************************************/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <sys/types.h>

#define	EXTERN extern
#include "llpam.h"
#include "token.h"

#define	ARITH_ARG(i, x)	\
    { \
	if (TAG(x) == TagINT) \
	    i = INT2LONG(x); \
	else \
	    i = am_arith_arg(x); \
    }
#define	FUNC_ARG(p)	am_arith_arg(STORE(p))

#define	EQUAL 0
#define	BEFORE -1
#define	AFTER 1

/****************************************************************
	Subroutines
****************************************************************/
void am_fast_write(CELL x, int needs_quote, FILE *stream)
{
    int i, n, t;
    STOREPTR p;
    CELL func;

    AM_DEREF(x, t, x);
    switch (t) {
    case TagREF:
	fprintf(stream, "_%ld", PTR2INX(REF2PTR(x)) - PTR2INX(heap_top));
	break;
    case TagINT:
	fprintf(stream, "%ld", INT2LONG(x));
	break;
    case TagCON:
	if (needs_quote)
	    writeq_string(SYMBOL(CON2SYM(x)).print_name, stream);
	else
	    write_string(SYMBOL(CON2SYM(x)).print_name, stream);
	break;
    case TagRES:
	fprintf(stream, "resource(%ld)", RES2RESINX(x));
	break;
    case TagLIS:
	write_string("[", stream);
	for (;;) {
	    am_fast_write(CAR(x), needs_quote, stream); 
	    AM_DEREF(x, t, CDR(x));
	    if (TAG(x) == TagLIS) {
		write_string(",", stream);
		continue;
	    }
	    if (x == SYM_NIL) {  
		break;
	    }
	    write_string("|", stream);
	    am_fast_write(x, needs_quote, stream);
	    break;
	}
	write_string("]", stream);
	break;
    case TagSTR:
	p = STR2PTR(x);
	func = STORE(p);
	am_fast_write(func, needs_quote, stream);	
	n = SYMBOL(CON2SYM(func)).arity;
	if (n <= 0) {
	    break;
	}
	write_string("(", stream);			
	for (i = 1; i <= n; i++) {
	    am_fast_write(STORE(p+i), needs_quote, stream);	
	    if(i < n)
		write_string(",", stream);		
	}
	write_string(")", stream);
	break;
    default:
	fail = YES;
	break;
    }
}


LONG am_arith_arg(CELL x)
{
    int t;
    STOREPTR p;
    LONG result;
    CELL func;

    AM_DEREF(x, t, x);
    switch (t) {
      case TagINT:
	return(INT2LONG(x));
      case TagLIS:
	return(am_arith_arg(STORE(LIS2PTR(x))));
      case TagSTR:
	p = STR2PTR(x);
	func = STORE(p);
        if (func == SYM_PLUS1)
	  result = FUNC_ARG(p+1);
        else if (func == SYM_PLUS)
	  result = FUNC_ARG(p+1) + FUNC_ARG(p+2);
        else if (func == SYM_MINUS1)
	  result = - FUNC_ARG(p+1);
        else if (func == SYM_MINUS)
	  result = FUNC_ARG(p+1) - FUNC_ARG(p+2);
        else if (func == SYM_TIMES)
	  result = FUNC_ARG(p+1) * FUNC_ARG(p+2);
        else if (func == SYM_DIVIDE)
	  result = FUNC_ARG(p+1) / FUNC_ARG(p+2);
        else if (func == SYM_QUOT)
	  result = FUNC_ARG(p+1) / FUNC_ARG(p+2);
        else if (func == SYM_MOD)
	  result = FUNC_ARG(p+1) % FUNC_ARG(p+2);
        else if (func == SYM_SHIFTL)
	  result = FUNC_ARG(p+1) << FUNC_ARG(p+2);
        else if (func == SYM_SHIFTR)
	  result = FUNC_ARG(p+1) >> FUNC_ARG(p+2);
        else if (func == SYM_NEG)
	  result = ~ FUNC_ARG(p+1);
        else if (func == SYM_AND)
	  result = FUNC_ARG(p+1) & FUNC_ARG(p+2);
        else if (func == SYM_OR)
	  result = FUNC_ARG(p+1) | FUNC_ARG(p+2);
        else if (func == SYM_XOR)
	  result = FUNC_ARG(p+1) ^ FUNC_ARG(p+2);
        else if (func == SYM_ABS)
	  result = abs( FUNC_ARG(p+1) );
        else if (func == SYM_MAX)
	  result = max( FUNC_ARG(p+1), FUNC_ARG(p+2) );
        else if (func == SYM_MIN)
	  result = min( FUNC_ARG(p+1), FUNC_ARG(p+2) );
        else {
	    printf("Arithmetic error ");
	    am_print_term(x);
	    printf("\n");
	    am_abort(3);
	}
	break;
      default:
	printf("Arithmetic error ");
	am_print_term(x);
	printf("\n");
	am_abort(4);
    }
    return(result);
}


int am_term_equal(CELL x, CELL y)
{
    STOREPTR p, q;
    int i, t1, t2;

    if (x == y)
      return(YES);
    AM_DEREF(x, t1, x);
    AM_DEREF(y, t2, y);
    if (t1 != t2)
      return(NO);
    switch (t1) {
      case TagREF:
      case TagINT:
      case TagCON:
      case TagRES:
	return(x == y);
      case TagLIS:
	p = LIS2PTR(x);
	q = LIS2PTR(y);
	for (i = 0; i < 2; i++)
	  if (! am_term_equal(STORE(p+i), STORE(q+i)))
	    return(NO);
	return(YES);
      case TagSTR:
	p = STR2PTR(x);
	q = STR2PTR(y);
	if (STORE(p) != STORE(q))
	  return(NO);
	for (i = 1; i <= SYMBOL(CON2SYM(STORE(p))).arity; i++)
	  if (! am_term_equal(STORE(p+i), STORE(q+i)))
	    return(NO);
	return(YES);
	break;
    }
    return(NO);
}


int am_make_structure(CELL x1, CELL x2, CELL x3)
{
    int t1, t2, t3;
    LONG arity, i;
    char *name;

    AM_DEREF(x1, t1, x1);
    AM_DEREF(x2, t2, x2);
    AM_DEREF(x3, t3, x3);
    if (t1 != TagREF || t3 != TagINT)
      return NO;
    arity = INT2LONG(x3);
    switch (t2) {
      case TagINT:
	if (arity != 0)
	  return NO;
	AM_ASSIGN(x1, x2);
	break;
      case TagCON:
	if (arity < 0 || 255 < arity)
	  return NO;
	name = SYMBOL(CON2SYM(x2)).print_name;
	if (strcmp(name, ".") == 0 && arity == 2) {
	    AM_ASSIGN(x1, MAKELIS(H));
	} else {
	    AM_ASSIGN(x1, MAKESTR(H));
	    STORE(H++) = MAKECON(am_make_symbol(name, arity));
	}
	for (i = 0; i < arity; i++) {
	    STORE(H) = MAKEREF(H);
	    H++;
	}
	break;
      default:
	return NO;
    }    
    return YES;
}


CELL am_string_chars(char *string)
{
    CELL list;
    char *p;

    if (*string == '\0') {
	list = SYM_NIL;
    } else {
	list = MAKELIS(H);
	for (p = string; *p; ++p) {
	    STORE(H  ) = MAKEINT(*p);
	    STORE(H+1) = MAKELIS(H+2);
	    H += 2;
	}
	STORE(H-1) = SYM_NIL;
    }
    return list;
}


int am_chars_string(CELL list, char *string)
{
    CELL x1, x2;
    int t1, t2;
    LONG code;
    char *p;

    AM_DEREF(x1, t1, list);
    p = string;
    while (LISTQ(x1)) {
	AM_DEREF(x2, t2, CAR(x1));
	if (t2 != TagINT)
	  return NO;
	code = INT2LONG(x2);
	if (code < 0 || 255 < code)
	  return NO;
	*p++ = code;
	AM_DEREF(x1, t1, CDR(x1));
    }
    if (x1 != SYM_NIL)
      return NO;
    *p = '\0';
    return YES;
}


#define MAXVARS 10000
struct {
    CELL old;
    CELL new;
} vars[MAXVARS];
int n_vars;

CELL am_copy_term1(CELL x)
{
    int t, i, n, ground;
    STOREPTR p;
    CELL a[MAXARGS];

    AM_DEREF(x, t, x);
    switch (t) {
    case TagREF:
	for (i = 0; i < n_vars; i++) {
	    if (x == vars[i].old)
		return vars[i].new;
	}
	vars[n_vars].old = x;
	x = vars[n_vars].new = STORE(H) = MAKEREF(H);
	n_vars++;
	H++;
	if (n_vars >= MAXVARS) {
	    printf("Copy term overflow\n");
	    am_abort(10);
	}
	return x;
    case TagLIS:
	p = LIS2PTR(x);
	ground = YES;
	for (i = 0; i <= 1; i++) {
	    a[i] = am_copy_term1(STORE(p+i));
	    if (a[i] == UNDEF)
		a[i] = STORE(p+i);
	    else
		ground = NO;
	}
	if (ground)
	    return UNDEF;
	p = H;
	for (i = 0; i <= 1; i++) {
	    STORE(H++) = a[i];
	}
	return MAKELIS(p);
    case TagSTR:
	p = STR2PTR(x);
	n = SYMBOL(CON2SYM(STORE(p))).arity;
	ground = YES;
	a[0] = STORE(p);
	for (i = 1; i <= n; i++) {
	    a[i] = am_copy_term1(STORE(p+i));
	    if (a[i] == UNDEF)
		a[i] = STORE(p+i);
	    else
		ground = NO;
	}
	if (ground)
	    return UNDEF;
	p = H;
	for (i = 0; i <= n; i++) {
	    STORE(H++) = a[i];
	}
	return MAKESTR(p);
    default:
	return UNDEF;
    }
}


CELL am_copy_term(CELL x)
{
    CELL z;

    n_vars = 0;
    z = am_copy_term1(x);
    if (z == UNDEF)
	return x;
    return z;
}


int am_ground(CELL x)
{
    int t, n, i;
    STOREPTR p;

    AM_DEREF(x, t, x);
    switch (t) {
    case TagREF:
        return NO;
    case TagLIS:
        p = LIS2PTR(x);
	return am_ground(STORE(p)) && am_ground(STORE(p+1));
    case TagSTR:
	p = STR2PTR(x);
	n = SYMBOL(CON2SYM(STORE(p))).arity;
	for (i = 1; i <= n; i++) {
	  if (! am_ground(STORE(p+i)))
	      return NO;
	}
	return YES;
    default:
        return YES;
    }
}


/*
  Variables < Integers < Atoms < Compound terms(containing lists)
          0      --> x = y
	  < 0    --> x < y
	  > 0    --> x > y

*/
int am_compare_term(CELL x, CELL y)
{
    STOREPTR p, q;
    int t1, t2;
    int i, result;
    int a1, a2;

    if (am_term_equal(x, y))
	return EQUAL;
    AM_DEREF(x, t1, x);
    AM_DEREF(y, t2, y);
    if (t1 == TagREF){
	if (t2 != TagREF)
	    return BEFORE;
	if ( x < y ) 
	    return BEFORE;
	return AFTER;
    }
    if (t1 == TagINT){
	if (t2 == TagREF) 
	    return AFTER;
	if (t2 != TagINT) 
	    return BEFORE;
	return (x - y);
    }
    if (t1 == TagCON) { 
	if (t2 == TagREF || t2 == TagINT)
	    return AFTER;
	if (t2 != TagCON) 
	    return BEFORE;
	return strcmp(SYMBOL(CON2SYM(x)).print_name, SYMBOL(CON2SYM(y)).print_name);
    }
    if (t1 == TagLIS) {
	if (t2 == TagREF || t2 == TagINT || t2 == TagCON)
	    return AFTER;
	if (t2 == TagSTR) { 
	    q = STR2PTR(y);
	    a2 = SYMBOL(CON2SYM(STORE(q))).arity;
	    if (a2 != 2) 
		return (2 - a2);
	    if (SYM_DOT != STORE(q))
		return am_compare_term(SYM_DOT, STORE(q));
	    p = LIS2PTR(x);
	    for (i = 0; i < 2; i++){
		result = am_compare_term(STORE(p+i), STORE(q+1+i));
		if (result != EQUAL)
		    return result;
	    }
	    return EQUAL;
	}
	if (t2 != TagLIS)
	    return BEFORE;
	p = LIS2PTR(x);
	q = LIS2PTR(y);
	for (i = 0; i < 2; i++){
	    result = am_compare_term(STORE(p+i), STORE(q+i));
	    if (result != EQUAL)
		return result;
	}
	return EQUAL;
    }
    if (t1 == TagSTR) {
	if (t2 == TagREF || t2 == TagINT || t2 == TagCON)
	    return AFTER;
	p = STR2PTR(x);
	a1 = SYMBOL(CON2SYM(STORE(p))).arity;
	if (t2 == TagLIS) {
	    if (a1 != 2) 
		return (a1 - 2);
	    if (STORE(p) != SYM_DOT)
		return am_compare_term(STORE(p), SYM_DOT);
	    q = LIS2PTR(y);
	    for (i = 0; i < 2; i++){
		result = am_compare_term(STORE(p+1+i), STORE(q+i));
		if (result != EQUAL)
		    return result;
	    }
	    return EQUAL;
	}
	if (t2 != TagSTR)
	    return BEFORE;
	q = STR2PTR(y);
	a2 = SYMBOL(CON2SYM(STORE(q))).arity;
	if (a1 != a2)
	    return (a1 - a2);
	if (STORE(p) != STORE(q))
	    return am_compare_term(STORE(p), STORE(q));
	for (i = 1; i <= a1 ; i++) {
	    result = am_compare_term(STORE(p+i), STORE(q+i));
	    if (result != EQUAL)
		return result;
	}
	return EQUAL;
    }
    return AFTER;
}


/****************************************************************
	System
****************************************************************/
void am_bi_statistics()
{
    static long old_runtime = 0;
    LONG val1, val2;

    switch (CARG2) {
      case 1:			/* runtime */
	val1 = usertime() - runtime0;
	val2 = val1 - old_runtime;
	old_runtime = val1;
	break;
      case 2:			/* code */
	val1 = sizeof(CELL) * (code_bot - code_top);
	val2 = sizeof(CELL) * CODESIZE - val1;
	break;
      case 3:			/* heap */
	val1 = sizeof(CELL) * (H - heap_top);
	val2 = sizeof(CELL) * HEAPSIZE - val1;
	break;
      case 4:			/* stack */
	val1 = sizeof(CELL) * (AM_STACK_BOT - stack_top);
	val2 = sizeof(CELL) * STACKSIZE - val1;
	break;
      case 5:			/* trail */
	val1 = sizeof(CELL) * (TR - trail_top);
	val2 = sizeof(CELL) * TRAILSIZE - val1;
	break;
      case 6:			/* pdl */
	val1 = sizeof(CELL) * 0;
	val2 = sizeof(CELL) * PDLSIZE - val1;
	break;
      case 7:			/* symbol */
	val1 = sizeof(sym_struct) * symbol_bot;
	val2 = sizeof(sym_struct) * SYMSIZE - val1;
	break;
      case 8:			/* resource */
	val1 = sizeof(res_struct) * R;
	val2 = sizeof(res_struct) * RESSIZE - val1;
	break;
      case 9:			/* hash, hash2 */
	val1 = sizeof(CELL) * (2*HASHSIZE);
	val2 = sizeof(CELL) * (2*HASHSIZE) - val1;
	break;
      case 10:			/* max_usage */
	val1 = max_usage;
	val2 = totalsize - max_usage;
	break;
      default:
	fail = YES;
	return;
    }
    am_unify(A(CARG3), MAKEINT(val1));
    if (fail)
      return;
    am_unify(A(CARG4), MAKEINT(val2));
    if (fail)
      return;
    P += 2;
}


void am_bi_load()
{
    CELL x;
    int t;

    AM_DEREF(x, t, A(CARG2));
    if (t != TagCON) {
	printf("File name error\n");
	fail = YES;
	return;
    }
    if (am_load(SYMBOL(CON2SYM(x)).print_name, 0)) {
	fail = YES;
	return;
    }
    CP = P + 1;
    P = load_base;
}


void am_bi_unix()
{
    CELL x;
    int t, rc;

    switch (CARG2) {
      case 1:			/* system */
	AM_DEREF(x, t, A(CARG3));
	if (t != TagCON) {
	    printf("System command error\n");
	    fail = YES;
	    return;
	}
	rc = system(SYMBOL(CON2SYM(x)).print_name);
	am_unify(A(CARG4), MAKEINT(rc));
	if (fail)
	  return;
	break;
      default:
	fail = YES;
	return;
    }
    P += 2;
}


/****************************************************************
	Input/Output
****************************************************************/
void am_bi_display()
{
    am_print_term(A(CARG2));
    P++;
}


void am_bi_nl()
{
    fprintf(stdout, "\n");
    if (isatty(fileno(stdout)))
      fflush(stdout);
    P++;
}


void am_bi_flush()
{
    fflush(stdout);
    P++;
}


void am_bi_put()
{
    CELL x;
    int t;

    x = A(CARG2);
    AM_DEREF(x, t, x);
    if (t != TagINT) {
	fail = YES;
	return;
    }
    fprintf(stdout, "%c", (int)INT2LONG(x));
    P++;
}


void am_bi_get0()
{
    long c;
    CELL x;

    x = A(CARG2);
    if (feof(stdin) && ! isatty(fileno(stdin))) {
	printf("Attempt to read past end of file\n");
	am_abort(9);
    }
    c = getchar();
    if (c == EOF)
      c = -1;
    am_unify(x, MAKEINT(c));
    if (fail)
      return;
    P++;
}


void am_bi_read_token()
{
    CELL x1, x2, a2;
    char *p;
    char string[BUFSIZ];
    int type;

    x1 = A(CARG2);
    x2 = A(CARG3);
    type = read_token(string, stdin);
    am_unify(x1, MAKEINT(type));
    if (fail)
      return;
    switch (type) {
      case 'I':
	a2 = MAKEINT(atol(string));
	break;
      case 'S':
	for (p = string; *p; p++) {
	}
	a2 = MAKECON(am_make_symbol(string, 0));
	break;
      default:
	a2 = MAKECON(am_make_symbol(string, 0));
	break;
    }
    am_unify(x2, a2);
    if (fail)
      return;
    P += 2;
}


void am_bi_fast_write()
{
    am_fast_write(A(CARG2), 0, stdout);
    P++;
}


void am_bi_fast_writeq()
{
    am_fast_write(A(CARG2), 1, stdout);
    P++;
}


void am_bi_atom_type()
{
    CELL x1, x2;
    int t1, t2;

    x1 = A(CARG2);
    x2 = A(CARG3);
    AM_DEREF(x1, t1, x1);
    if (t1 != TagCON) {
	fail = YES;
	return;
    }
    am_unify(x2, MAKEINT(string_type(SYMBOL(CON2SYM(x1)).print_name)));
    P += 2;
}


/****************************************************************
	Meta Logic
****************************************************************/
void am_bi_var()
{
    CELL x;
    int t;

    AM_DEREF(x, t, A(CARG2));
    if (t != TagREF) {
	fail = YES;
	return;
    }
    P++;
}


void am_bi_nonvar()
{
    CELL x;
    int t;

    AM_DEREF(x, t, A(CARG2));
    if (t == TagREF) {
	fail = YES;
	return;
    }
    P++;
}


void am_bi_atomic()
{
    CELL x;
    int t;

    AM_DEREF(x, t, A(CARG2));
    if (t != TagCON && t != TagINT) {
	fail = YES;
	return;
    }
    P++;
}


void am_bi_integer()
{
    CELL x;
    int t;

    AM_DEREF(x, t, A(CARG2));
    if (t != TagINT) {
	fail = YES;
	return;
    }
    P++;
}


void am_bi_atom()
{
    CELL x;
    int t;

    AM_DEREF(x, t, A(CARG2));
    if (t != TagCON) {
	fail = YES;
	return;
    }
    P++;
}


void am_bi_functor()
{
    CELL x1, x2, x3;
    int t1;
    CELL functor;
    int arity;

    AM_DEREF(x1, t1, A(CARG2));
    x2 = A(CARG3);
    x3 = A(CARG4);
    if (t1 == TagREF) {
	if (am_make_structure(x1, x2, x3) != YES) {
	    printf("Error in functor(");
	    am_print_term(A(CARG2));
	    printf(",");
	    am_print_term(A(CARG3));
	    printf(",");
	    am_print_term(A(CARG4));
	    printf(")\n");
	    am_abort(5);
	}
    } else {
	switch (t1) {
	  case TagINT:
	  case TagCON:
	    functor = x1;
	    arity = 0;
	    break;
	  case TagSTR:
	    functor = STORE(STR2PTR(x1));
	    arity = SYMBOL(CON2SYM(functor)).arity;
	    functor = MAKECON(am_make_symbol(
				    SYMBOL(CON2SYM(functor)).print_name, 0));
	    break;
	  case TagLIS:
	    functor = SYM_DOT;
	    arity = 2;
	    break;
	  default:
	    AM_FAIL;
	    return;
	}
	am_unify(x2, functor);
	if (fail)
	  return;
	am_unify(x3, MAKEINT(arity));
    }
    P += 2;
}


void am_bi_arg()
{
    CELL x1, x2, x3;
    int t1, t2, t3;
    STOREPTR p;
    LONG arity, i;

    AM_DEREF(x1, t1, A(CARG2));
    AM_DEREF(x2, t2, A(CARG3));
    AM_DEREF(x3, t3, A(CARG4));
    if (t1 != TagINT || t2 == TagREF) {
	printf("Error in arg(");
	am_print_term(A(CARG2));
	printf(",");
	am_print_term(A(CARG3));
	printf(",");
	am_print_term(A(CARG4));
	printf(")\n");
	am_abort(6);
    }
    switch (t2) {
      case TagINT:
      case TagCON:
      case TagRES:
	fail = YES;
	return;
      case TagSTR:
	p = STR2PTR(x2) + 1;
	arity = SYMBOL(CON2SYM(STORE(STR2PTR(x2)))).arity;
	break;
      case TagLIS:
	p = STR2PTR(x2);
	arity = 2;
	break;
    }
    i = INT2LONG(x1);
    if (i <= 0 || arity < i) {
	fail = YES;
	return;
    }
    am_unify(x3, STORE(p + i - 1));
    P += 2;
}


void am_bi_atom_chars()
{
    CELL x1, x2;
    int t1, t2;
    CELL list;
    char string[BUFSIZ];

    AM_DEREF(x1, t1, A(CARG2));
    AM_DEREF(x2, t2, A(CARG3));
    if (t1 == TagCON) {
	list = am_string_chars(SYMBOL(CON2SYM(x1)).print_name);
	am_unify(list, x2);
    } else if (t1 == TagREF && am_chars_string(x2, string) == YES) {
	AM_ASSIGN(x1, MAKECON(am_make_symbol(string, 0)));
    } else {
	printf("Error in atom_chars(");
	am_print_term(A(CARG2));
	printf(",");
	am_print_term(A(CARG3));
	printf(")\n");
	am_abort(7);
    }
    P += 2;
}


void am_bi_number_chars()
{
    CELL x1, x2;
    int t1, t2;
    CELL list;
    char string[BUFSIZ];
    char *p;

    AM_DEREF(x1, t1, A(CARG2));
    AM_DEREF(x2, t2, A(CARG3));
    if (t1 == TagINT) {
	sprintf(string, "%ld", INT2LONG(x1));
	list = am_string_chars(string);
	am_unify(list, x2);
    } else if (t1 == TagREF && am_chars_string(x2, string) == YES) {
	p = string;
	if (*p == '-')
	  p++;
	for (; *p; p++) {
	    if (! isdigit(*p)) {
		fail = YES;
		return;
	    }
	}
	AM_ASSIGN(x1, MAKEINT(atol(string)));
    } else {
	printf("Error in number_chars(");
	am_print_term(A(CARG2));
	printf(",");
	am_print_term(A(CARG3));
	printf(")\n");
	am_abort(8);
    }
    P += 2;
}


void am_bi_copy_term()
{
    CELL x;
    x = am_copy_term(A(CARG2));
    am_unify(A(CARG3), x);
    P += 2;
}


void am_bi_ground()
{
    if (! am_ground(A(CARG2))) {
        fail = YES;
	return;
    }
    P++;
}


/****************************************************************
	Arithmetic
****************************************************************/
void am_bi_arith()
{
    int op;
    LONG a1, a2, result;
    CELL x1, x2, x3;
    int t1, t2, t3;

    op = CARG2;
    AM_DEREF(x1, t1, A(CARG3));
    ARITH_ARG(a1, x1);
    if (CARG4 > 0) {
        AM_DEREF(x2, t2, A(CARG4));
	ARITH_ARG(a2, x2);
    }
    x3 = A(CARG5);
    switch (op) {
      case  0: result =  a1 +  a2 ; break;
      case  1: result =  a1 -  a2 ; break;
      case  2: result =  a1 *  a2 ; break;
      case  3: result =  a1 /  a2 ; break;
      case  4: result =  a1 %  a2 ; break;
      case  5: result =  a1 << a2 ; break;
      case  6: result =  a1 >> a2 ; break;
      case  7: result =  - a1     ; break;
      case 10: result =  ~ a1     ; break;
      case 11: result =  a1 &  a2 ; break;
      case 12: result =  a1 |  a2 ; break;
      case 13: result =  a1 ^  a2 ; break;
      case 20: result =  abs(a1)  ; break;
      case 21: result = max(a1,a2); break;
      case 22: result = min(a1,a2); break;
      case 50: result = a1        ; break;
      default: fail = YES; return;
    }
    AM_DEREF(x3, t3, x3);
    switch (t3) {
      case TagREF:
        AM_ASSIGN(x3, MAKEINT(result));
	break;
      case TagINT:
	if (INT2LONG(x3) != result) {
	    fail = YES;
	    return;
	}
	break;
      default:
	fail = YES;
	return;
    }
    P += 2;
}


/****************************************************************
	Comparison
****************************************************************/
void am_bi_compare()
{
    int op, ok;
    CELL x1, x2;
    LONG a1, a2;
    int t1, t2;

    op = CARG2;
    AM_DEREF(x1, t1, A(CARG3));
    AM_DEREF(x2, t2, A(CARG4));
    switch (op) {
    case 1:  /* == */
	ok = am_term_equal(x1, x2);
	break;
    case 2:  /* \== */
	ok = ! am_term_equal(x1, x2);
	break;
    case 3:  /* @> */
	ok =  (am_compare_term(x1, x2) > 0);
	break;
    case 4:  /* @=< */
	ok =  (am_compare_term(x1, x2) <= 0);
	break;
    case 10:  /* =:= */
        ARITH_ARG(a1, x1);
        ARITH_ARG(a2, x2);
	ok = (a1 == a2);
	break;
    case 11:  /*  =\= */
        ARITH_ARG(a1, x1);
        ARITH_ARG(a2, x2);
	ok = (a1 != a2);
	break;
    case 12:  /* > */
        ARITH_ARG(a1, x1);
        ARITH_ARG(a2, x2);
	ok = (a1 >  a2);
	break;
    case 13:  /* <= */
        ARITH_ARG(a1, x1);
        ARITH_ARG(a2, x2);
	ok = (a1 <= a2);
	break;
    default:
	fail = YES;
	return;
    }
    if (! ok) {
	fail = YES;
	return;
    }
    P += 2;
}


void am_bi_equal()
{
    am_unify(A(CARG2), A(CARG3));
    if (fail) {
	return;
    }
    P += 2;
}


/****************************************************************
	Occurs Check
****************************************************************/
#define ACYC_HASHSIZE (521)
static CELL phash[ACYC_HASHSIZE];
static CELL vhash[ACYC_HASHSIZE];

int am_term_acyc(CELL x)
{
    STOREPTR p;
    CELL d, q;
    int t, i, n, h;

    AM_DEREF(d, t, x);
    switch (t) {
      case TagREF:
      case TagINT:
      case TagCON:
      case TagRES:
	  break;
      case TagLIS:
      case TagSTR:
	  h = (ULONG)d % ACYC_HASHSIZE;
	  for (q = phash[h]; LISTQ(q); q = CDR(q)) {
	      if (d == STORE(LIS2PTR(q))) {
		  return NO;
	      }
	  }
	  for (q = vhash[h]; LISTQ(q); q = CDR(q)) {
	      if (d == STORE(LIS2PTR(q))) {
		  return YES;
	      }
	  }
	  if (H + MARGIN >= stack_top) {
	      printf(" HEAP area (%ld) overflow\n", HEAPSIZE);
	      am_abort(102);
	  }
	  STORE(H) = d;
	  STORE(H+1) = phash[h];
	  phash[h] = MAKELIS(H);
	  H += 2;
	  STORE(H) = d;
	  STORE(H+1) = vhash[h];
	  vhash[h] = MAKELIS(H);
	  H += 2;
	  if (t == TagLIS) {
	      p = LIS2PTR(d);
	      if (! am_term_acyc(STORE(p)))
		  return NO;
	      if (! am_term_acyc(STORE(p+1)))
		  return NO;
	  } else {
	      p = STR2PTR(d);
	      n = SYMBOL(CON2SYM(STORE(p))).arity;
	      for (i = 1; i <= n; i++) {
		  if (! am_term_acyc(STORE(p+i)))
		      return NO;
	      }
	  }
	  phash[h] = CDR(phash[h]);
	  break;
    }
    return YES;
}


void am_term_acyclic(CELL x)
{
    STOREPTR oldH;
    int i;

    oldH = H;
    for (i = 0; i < ACYC_HASHSIZE; i++) {
	phash[i] = SYM_NIL;
	vhash[i] = SYM_NIL;
    }
    if (! am_term_acyc(x))
	fail = YES;
    H = oldH;
}


void am_unify_with_occurs_check(CELL a1, CELL a2)
{
    STOREPTR pdl, p1, p2;
    CELL d1, d2;
    int t1, t2, i, n;

    pdl = pdl_bot;
    *--pdl = a1;
    *--pdl = a2;
    while (!fail && pdl < pdl_bot) {
	AM_DEREF(d2, t2, *pdl++);
	AM_DEREF(d1, t1, *pdl++);
	if (d1 == d2) {
	    continue;
	}
	if (t1 == TagREF || t2 == TagREF) {
	    if (t1 == TagREF && (t2 != TagREF || d2 < d1)) {
		if (am_occurs_check(d1, d2)) {
		    fail = YES;
		    continue;
		}
		AM_ASSIGN(d1, d2);
		/*
		am_term_acyclic(d1);
		*/
	    }
	    else {
		if (am_occurs_check(d2, d1)) {
		    fail = YES;
		    continue;
		}
		AM_ASSIGN(d2, d1);
		/*
		am_term_acyclic(d2);
		*/
	    }
	    continue;
	}
	if (t1 != t2) {
	    fail = YES;
	    continue;
	}
	switch (t1) {
	  case TagINT:
	  case TagCON:
	  case TagRES:
	    fail = (d1 != d2);
	    break;
	  case TagLIS:
	    p1 = LIS2PTR(d1);
	    p2 = LIS2PTR(d2);
	    *--pdl = STORE(p1);
	    *--pdl = STORE(p2);
	    *--pdl = STORE(p1+1);
	    *--pdl = STORE(p2+1);
	    if (pdl - MARGIN <= pdl_top) {
		printf("PDL area (%ld) overflow\n", PDLSIZE);
		am_abort(107);
	    }
	    break;
	  case TagSTR:
	    p1 = STR2PTR(d1);
	    p2 = STR2PTR(d2);
	    if (STORE(p1) != STORE(p2)) {
		fail = YES;
		break;
	    }
	    n = SYMBOL(CON2SYM(STORE(p1))).arity;
	    for (i = 1; i <= n; i++) {
		*--pdl = STORE(p1+i);
		*--pdl = STORE(p2+i);
	    }
	    if (pdl - MARGIN <= pdl_top) {
		printf("PDL area (%ld) overflow\n", PDLSIZE);
		am_abort(106);
	    }
	    break;
	}
    }
}


int am_check_unify(int dep, CELL a1, CELL a2)
{
    STOREPTR p1, p2;
    CELL d1, d2;
    int t1, t2, i, n;

    if (dep <= 0)
	return YES;
    AM_DEREF(d1, t1, a1);
    AM_DEREF(d2, t2, a2);
    if (d1 == d2)
	return YES;
    if (t1 == TagREF || t2 == TagREF)
	return YES;
    if (t1 != t2)
	return NO;
    switch (t1) {
    case TagINT:
    case TagCON:
    case TagRES:
	return d1 == d2;
    case TagLIS:
	p1 = LIS2PTR(d1);
	p2 = LIS2PTR(d2);
	return am_check_unify(dep-1, STORE(p1), STORE(p2))
	    && am_check_unify(dep-1, STORE(p1+1), STORE(p2+1));
    case TagSTR:
	p1 = STR2PTR(d1);
	p2 = STR2PTR(d2);
	if (STORE(p1) != STORE(p2))
	    return NO;
	n = SYMBOL(CON2SYM(STORE(p1))).arity;
	for (i = 1; i <= n; i++) {
	    if (! am_check_unify(dep-1, STORE(p1+i), STORE(p2+i)))
		return NO;
	}
	return YES;
    }
    return NO;
}


void am_bi_unify_with_occurs_check()
{
    CELL d1, d2;
    int t1, t2, i;

    AM_DEREF(d1, t1, A(CARG2));
    AM_DEREF(d2, t2, A(CARG3));
    if (! am_check_unify(4, d1, d2)) {
	fail = YES;
	return;
    }
    am_term_acyclic(d1);
    if (fail)
	return;
    am_term_acyclic(d2);
    if (fail)
	return;
    am_unify_with_occurs_check(d1, d2);
    P += 2;
}


void am_bi_term_acyclic()
{
    am_term_acyclic(A(CARG2));
    P++;
}


/****************************************************************
	prolog_flag
****************************************************************/
void am_bi_prolog_flag()
{
    CELL d;
    int x, t;

    x = CARG2;
    AM_DEREF(d, t, A(CARG3));
    P += 2;
    switch (x) {
    case 0:	/* unknown */
	if (t == TagREF) {
	    switch (prolog_flag.unknown) {
	    case 0:
		AM_ASSIGN(d, MAKECON(am_make_symbol("error",  0)));
		return;
	    case 1:
		AM_ASSIGN(d, MAKECON(am_make_symbol("fail",  0)));
		return;
	    }
	} else if (d == MAKECON(am_make_symbol("error",  0))) {
	    prolog_flag.unknown = 0;
	    return;
	} else if (d == MAKECON(am_make_symbol("fail",  0))) {
	    prolog_flag.unknown = 1;
	    return;
	}
	printf("Illegal prolog_flag for unknown\n");
	break;
    case 1:	/* occurs_check */
	if (t == TagREF) {
	    switch (prolog_flag.occurs_check) {
	    case 0:
		AM_ASSIGN(d, MAKECON(am_make_symbol("off",  0)));
		return;
	    case 1:
		AM_ASSIGN(d, MAKECON(am_make_symbol("on",  0)));
		return;
	    }
	} else if (d == MAKECON(am_make_symbol("off",  0))) {
	    prolog_flag.occurs_check = 0;
	    return;
	} else if (d == MAKECON(am_make_symbol("on",  0))) {
	    prolog_flag.occurs_check = 1;
	    return;
	}
	printf("Illegal prolog_flag for occurs_check\n");
	break;
    case 2:	/* max_usage */
	if (t == TagREF) {
	    switch (prolog_flag.max_usage) {
	    case 0:
		AM_ASSIGN(d, MAKECON(am_make_symbol("off",  0)));
		return;
	    case 1:
		AM_ASSIGN(d, MAKECON(am_make_symbol("on",  0)));
		return;
	    }
	} else if (d == MAKECON(am_make_symbol("off",  0))) {
	    prolog_flag.max_usage = 0;
	    return;
	} else if (d == MAKECON(am_make_symbol("on",  0))) {
	    prolog_flag.max_usage = 1;
	    return;
	}
	printf("Illegal prolog_flag for max_usage\n");
	break;
    default:
	printf("Illegal prolog_flag\n");
	break;
    }
    fail = YES;
}


/****************************************************************
	Invoke Built-in Predicates
****************************************************************/
typedef void (*builtin_func_ptr)();
builtin_func_ptr builtin_dispatch[] = {
    am_bi_statistics,
    am_bi_load,
    am_bi_unix,
    am_bi_display,
    am_bi_nl,
    am_bi_flush,
    am_bi_put,
    am_bi_get0,
    am_bi_read_token,
    am_bi_fast_write,
    am_bi_fast_writeq,
    am_bi_atom_type,
    am_bi_var,
    am_bi_nonvar,
    am_bi_atomic,
    am_bi_integer,
    am_bi_atom,
    am_bi_functor,
    am_bi_arg,
    am_bi_atom_chars,
    am_bi_number_chars,
    am_bi_copy_term,
    am_bi_ground,
    am_bi_arith,
    am_bi_compare,
    am_bi_equal,
    am_bi_unify_with_occurs_check,
    am_bi_term_acyclic,
    am_bi_prolog_flag
};
#define	MAXBUILTIN	(sizeof(builtin_dispatch)/sizeof(builtin_func_ptr))

/* BUILTIN1 & BUILTIN2 */
void am_builtin()
{
    CELL x1;
    int func_no;
    builtin_func_ptr fp;

    func_no = CARG1;
    if (func_no < 0 || func_no >= MAXBUILTIN) {
	printf("Illegal builtin function #%d\n", func_no);
	am_abort(310);
    }
    fp = builtin_dispatch[func_no];
    (*fp)();
    if (fail) {
	AM_FAIL;
    }
}

/*
 Local Variables:
 mode: c
 c-indent-level: 4
 End:
*/
