/****************************************************************
	LLPAM Instruction Table
	by Naoyuki Tamura (tamura@kobe-u.ac.jp)
	   Dept. CS, Kobe University

	Last modified by Mutsunori Banbara on 29 July 2000
	(banbara@cs32.scitec.kobe-u.ac.jp)
****************************************************************/
/*
	Type	Format
	 0	OP
	 1	OP+ARG1
	 2	OP+ARG1+ARG2
	 3	OP+ARG1, f/n
	 4	OP+ARG1, c
	 5	OP+ARG1, i
	 6	OP, f/n
	 7	OP, c
	 8	OP, i
	 9	OP, L
	10	OP, L1,L2,L3,L4,L5
	11	OP+ARG1+ARG2, ARG3
	12	OP+ARG1+ARG2, ARG3+ARG4
	13	OP+ARG1+ARG2, ARG3+ARG4+ARG5
	14	OP+ARG1, Term
	15	OP, Term
	16	OP+ARG1, L, i
	17	OP+ARG1, f/n, L
	18	OP+ARG1+ARG2, L
	19      OP+N, L1,...,LN
*/
enum INSTR {
    HALT,
    FAIL,
    TRACE,
    NOP,
    DEREF_A,
    DEREF_Y,
    BUILTIN1,
    BUILTIN2,
    DEBUG_DMP_A,
    DEBUG_DMP_Y,
    DEBUG_PRN_A,
    DEBUG_PRN_Y,
    DEBUG_DMP_E,
    DEBUG_DMP_B,
    DEBUG_DMP_S,
    DEBUG_DMP_R,
    PUT_VAR_A,
    PUT_VAR_Y,
    PUT_VAL_A,
    PUT_VAL_Y,
    PUT_UNSAFE_VAL,
    PUT_STR,
    PUT_LIST,
    PUT_CON,
    PUT_INT,
    PUT_GROUND,
    GET_VAR_A,
    GET_VAR_Y,
    GET_VAL_A,
    GET_VAL_Y,
    GET_STR,
    GET_LIST,
    GET_CON,
    GET_INT,
    GET_GROUND,
    SET_VAR_A,
    SET_VAR_Y,
    SET_VAL_A,
    SET_VAL_Y,
    SET_LOCAL_VAL_A,
    SET_LOCAL_VAL_Y,
    SET_CON,
    SET_INT,
    SET_GROUND,
    SET_VOID,
    UNIFY_VAR_A,
    UNIFY_VAR_Y,
    UNIFY_VAL_A,
    UNIFY_VAL_Y,
    UNIFY_LOCAL_VAL_A,
    UNIFY_LOCAL_VAL_Y,
    UNIFY_CON,
    UNIFY_INT,
    UNIFY_GROUND,
    UNIFY_VOID,
    ALLOCATE,
    DEALLOCATE,
    CALL,
    EXECUTE,
    EXECUTE_INDIRECT,
    EXECUTE_INDIRECT_NOTRACE,
    EXECUTE_PROG_INDIRECT,
    PROCEED,
    TRY_ME_ELSE,
    RETRY_ME_ELSE,
    TRUST_ME,
    TRY,
    RETRY,
    TRUST,
    SWITCH_ON_TERM,
    SWITCH_ON_HASH,
    NECK_CUT,
    GET_LEVEL,
    CUT,
    BEGIN_IMP_A,
    BEGIN_IMP_Y,
    BEGIN_EXP_IMP_A,
    BEGIN_EXP_IMP_Y,
    ADD_RES,
    ADD_EXP_RES,
    MID_IMP_A,
    MID_IMP_Y,
    MID_EXP_IMP_A,
    MID_EXP_IMP_Y,
    END_IMP_A,
    END_IMP_Y,
    END_EXP_IMP_A,
    END_EXP_IMP_Y,
    MORE_IMP,
    BEGIN_WITH_A,
    BEGIN_WITH_Y,
    MID_WITH_A,
    MID_WITH_Y,
    END_WITH_A,
    END_WITH_Y,
    BEGIN_BANG_A,
    BEGIN_BANG_Y,
    END_BANG_A,
    END_BANG_Y,
    UNIFY_RESOURCE,
    RESOURCE_BODY,
    TOP,
    ERASE,
    DEBUG_DMP_R_REG,
    PUT_CLO,
    EXECUTE_CLO,
    PICKUP_RES,
    CONSUME,
    RESTORE_RES,
    TRY_RES,
    RETRY_RES_ELSE,
    TRUST_RES,
    IF_NO_RES,
    ADD_RES_TERM,
    ADD_EXP_RES_TERM,
    PICKUP_RES_TERM,
    EXECUTE_RES_INDIRECT,
    LAST_OP
};

#ifdef DECLARE_OP_TABLE
struct {
    char *op_name;
    int op_type;
} op_table[] = {
    {"halt",			0 },
    {"fail",			0 },
    {"trace",			1 },
    {"nop",			0 },
    {"deref_a",			1 },
    {"deref_y",			1 },
    {"builtin1",		2 },
    {"builtin2",		13 },
    {"debug_dmp_a",		1 },
    {"debug_dmp_y",		1 },
    {"debug_prn_a",		1 },
    {"debug_prn_y",		1 },
    {"debug_dmp_e",		0 },
    {"debug_dmp_b",		0 },
    {"debug_dmp_s",		0 },
    {"debug_dmp_r",		0 },
    {"put_var_a",		2 },
    {"put_var_y",		2 },
    {"put_val_a",		2 },
    {"put_val_y",		2 },
    {"put_unsafe_val",		2 },
    {"put_str",			3 },
    {"put_list",		1 },
    {"put_con",			4 },
    {"put_int",			5 },
    {"put_ground",		14 },
    {"get_var_a",		2 },
    {"get_var_y",		2 },
    {"get_val_a",		2 },
    {"get_val_y",		2 },
    {"get_str",			3 },
    {"get_list",		1 },
    {"get_con",			4 },
    {"get_int",			5 },
    {"get_ground",		14 },
    {"set_var_a",		1 },
    {"set_var_y",		1 },
    {"set_val_a",		1 },
    {"set_val_y",		1 },
    {"set_local_val_a",		1 },
    {"set_local_val_y",		1 },
    {"set_con",			7 },
    {"set_int",			8 },
    {"set_ground",		15 },
    {"set_void",		1 },
    {"unify_var_a",		1 },
    {"unify_var_y",		1 },
    {"unify_val_a",		1 },
    {"unify_val_y",		1 },
    {"unify_local_val_a",	1 },
    {"unify_local_val_y",	1 },
    {"unify_con",		7 },
    {"unify_int",		8 },
    {"unify_ground",		15 },
    {"unify_void",		1 },
    {"allocate",		1 },
    {"deallocate",		0 },
    {"call",			6 },
    {"execute",			6 },
    {"execute_indirect",	1 },
    {"execute_indirect_notrace",1 },
    {"execute_prog_indirect",	1 },
    {"proceed",			0 },
    {"try_me_else",		9 },
    {"retry_me_else",		9 },
    {"trust_me",		0 },
    {"try",			9 },
    {"retry",			9 },
    {"trust",			9 },
    {"switch_on_term",		10 },
    {"switch_on_hash",		19 },
    {"neck_cut",		0 },
    {"get_level",		1 },
    {"cut",			1 },
    {"begin_imp_a",		1 },
    {"begin_imp_y",		1 },
    {"begin_exp_imp_a",		1 },
    {"begin_exp_imp_y",		1 },
    {"add_res",			2 }, 
    {"add_exp_res",		2 },
    {"mid_imp_a",		2 },
    {"mid_imp_y",		2 },
    {"mid_exp_imp_a",		2 },
    {"mid_exp_imp_y",		2 },
    {"end_imp_a",		11 },
    {"end_imp_y",		11 },
    {"end_exp_imp_a",		11 },
    {"end_exp_imp_y",		11 },
    {"more_imp",		0 },
    {"begin_with_a",		1 },
    {"begin_with_y",		1 },
    {"mid_with_a",		1 },
    {"mid_with_y",		1 },
    {"end_with_a",		2 },
    {"end_with_y",		2 },
    {"begin_bang_a",		1 },
    {"begin_bang_y",		1 },
    {"end_bang_a",		1 },
    {"end_bang_y",		1 },
    {"unify_resource",		2 },
    {"resource_body",		2 },
    {"top",			0 },
    {"erase",			0 },
    {"debug_dmp_r_reg",		0 },
    {"put_closure",		16 },
    {"execute_closure",		1 },
    {"pickup_resource",		17 },
    {"consume", 		2 },
    {"restore_resource",	0 },
    {"try_resource",    	9 },
    {"retry_resource_else",	9 },
    {"trust_resource",  	9 },
    {"if_no_resource",		9 },
    {"add_res_term",		2 }, 
    {"add_exp_res_term",	2 },
    {"pickup_resource_term",	18 },
    {"execute_resource_indirect", 1 },
};
#endif

