/****************************************************************
	LLPAM Header File
	by Naoyuki Tamura (tamura@kobe-u.ac.jp)
	   Dept. CS, Kobe University
	Last modified by Mutsunori Banbara on Jaly. 27, 2000
	(banbara@cs32.scitec.kobe-u.ac.jp)
****************************************************************/
#include <setjmp.h>
#include "sysdep.h"

/****************************************************************
	Constants
****************************************************************/
#define Kword	(1024)
#define YES	(1)
#define NO	(0)
#define TRUE	(1)
#define FALSE	(0)

/****************************************************************
	Size of Data Areas
****************************************************************/
#define CODESIZE	( 100*Kword)	/* CODE */
#define HEAPSIZE	(1000*Kword)	/* HEAP */
#define STACKSIZE	(1000*Kword)	/* STACK */
#define TRAILSIZE	(1000*Kword)	/* TRAIL */
#define PDLSIZE		( 100*Kword)	/* PDL */
#define	MARGIN		(1*Kword)

#define SYMSIZE		(4000)		/* SYMBOL Table */
#define RESSIZE		(4000)		/* RES Table */
#define HASHSIZE	(2203)		/* HASH Table */

#define RESMARGIN	(100)
#define LEVELMARGIN	(100)

/****************************************************************
	STORE Areas
****************************************************************/
/* CELL (4 bytes), (CELL*) should have the same size */
typedef int BOOLEAN; 
typedef unsigned long int CELL;
typedef long int LONG;
typedef unsigned long int ULONG;
typedef unsigned char BYTE;
/* The following pointers are initialized in am_init */
EXTERN CELL *store;     	/* top of STORE (const) */
typedef CELL *STOREPTR;
EXTERN STOREPTR code_top;	/* top of CODE  (const) */
EXTERN STOREPTR code_bot;	/* bottom of CODE */
EXTERN STOREPTR heap_top;	/* top of HEAP  (const) */
EXTERN STOREPTR stack_top;	/* top of STACK (const) */
EXTERN STOREPTR trail_top;	/* top of TRAIL (const) */
EXTERN STOREPTR pdl_top;	/* top of PDL   (const) */
EXTERN STOREPTR pdl_bot;	/* bot of PDL   (const) */
/* Basic Access Methods */
#define	STORE(x)	(*(STOREPTR)(x))
#define	PTR2INX(p)	((p)-store)
#define	INX2PTR(i)	(store+(i))

/* CODE CELL
    8-bits: OP code, ARG3
    8-bits: for debugging 
    8-bits: ARG1, ARG4
    8-bits: ARG2, ARG5
*/
#define OP		(*P>>20)
#define CARG1		((*P<<12)>>22)
#define CARG2		((*P<<22)>>22)
#define CARG3		(*(P+1)>>20)
#define CARG4		((*(P+1)<<12)>>22)
#define CARG5		((*(P+1)<<22)>>22)
#define CARGW(i)	*(P+(i))

/* STORE CELL
    1-bit : GC mark bit
    3-bits: Tag field (REF, INT, CON, STR, LIS, RES, CLO)
   28-bits: Data field
*/
#define	TAG(x)		(int)((ULONG)(x)>>28)
#define	MASK(x)		(((ULONG)(x)<<4)>>4)
#define	DATA(x)		MASK(x)
#define	REF2PTR(x)	INX2PTR(x)
#define	INT2LONG(x)	(((LONG)(x)<<4)>>4)
#define	CON2SYM(x)	DATA(x)
#define	STR2PTR(x)	INX2PTR(DATA(x))
#define	LIS2PTR(x)	INX2PTR(DATA(x))
#define	RES2INX(x)	DATA(x)
#define	RES2RESINX(x)	DATA(x)
#define	MAKECELL(t,d)	(((CELL)(t)<<28)|(d))
#define	MAKEREF(p)	(CELL)PTR2INX(p)
#define	MAKEINT(i)	MAKECELL(TagINT, MASK(i))
#define	MAKECON(i)	MAKECELL(TagCON, MASK(i))
#define	MAKESTR(p)	MAKECELL(TagSTR, PTR2INX(p))
#define	MAKELIS(p)	MAKECELL(TagLIS, PTR2INX(p))
#define	MAKERES(i)	MAKECELL(TagRES, MASK(i))
#define	MAXINT		(0x07ffffffL)
#define	CLO2PTR(x)	INX2PTR(DATA(x))
#define	MAKECLO(p)	MAKECELL(TagCLO, PTR2INX(p))

#define	MAKETOP(i)	MAKECELL(TagTOP, MASK(i))
#define	TOP2BOOL(x)	(BOOLEAN)DATA(x)

/****************************************************************
	SYMBOL Table
****************************************************************/
typedef struct  {
    char *print_name;
    int arity;
    ULONG hash_value;
    STOREPTR codeaddr;
    int plevel;		/* privilege level of the predicate */
    CELL res;
    CELL res2;
    CELL res3;
    CELL res4;
} sym_struct;

EXTERN sym_struct *symbol;
typedef long int SYMBOLINX;
EXTERN SYMBOLINX symbol_bot;
#define SYMBOL(x)	symbol[x]
#define PRED_UNDEF	((STOREPTR)0)
#define	UNDEF		(CELL)(-1L)

/****************************************************************
	RES Table
****************************************************************/
typedef struct  {
    int s1;
    int s2;
    int level;
    int deadline;
    BOOLEAN out_of_scope;
    CELL head;
    CELL body;
    CELL closure;   
    SYMBOLINX pred; 
} res_struct;

EXTERN res_struct *res;
typedef long int RESINX;
#define RES(x)   res[x]

EXTERN STOREPTR hash;
#define HASH(x)	hash[x]

EXTERN STOREPTR hash2;
#define HASH2(x) hash2[x]


/****************************************************************
	Tags & Symbols
****************************************************************/
/* Tag values */
enum tag { 
    TagREF, 	/* should be 0 */
    TagINT, 	/* integer */
    TagCON, 	/* symbol */
    TagSTR, 	/* structure pointer */
    TagLIS,	/* list pointer */
    TagRES,     /* resource pointer */
    TagCLO,     /* closure */
    TagTOP
   };

/* Special symbols */
EXTERN CELL SYM_NIL;
EXTERN CELL SYM_DOT;
EXTERN CELL SYM_PLUS1;
EXTERN CELL SYM_PLUS;
EXTERN CELL SYM_MINUS1;
EXTERN CELL SYM_MINUS;
EXTERN CELL SYM_TIMES;
EXTERN CELL SYM_DIVIDE;
EXTERN CELL SYM_QUOT;
EXTERN CELL SYM_MOD;
EXTERN CELL SYM_SHIFTL;
EXTERN CELL SYM_SHIFTR;
EXTERN CELL SYM_NEG;
EXTERN CELL SYM_AND;
EXTERN CELL SYM_OR;
EXTERN CELL SYM_XOR;
EXTERN CELL SYM_ABS;
EXTERN CELL SYM_MAX;
EXTERN CELL SYM_MIN;

EXTERN CELL SYM_RES;
EXTERN CELL SYM_TRUE;
EXTERN CELL SYM_TRACE;

/******************************************************************
	LLPAM Registers & Flags
******************************************************************/
#ifdef USE_REGISTER
  register STOREPTR P  asm (REG_P);
  register STOREPTR S  asm (REG_S);
  register STOREPTR H  asm (REG_H);
  EXTERN STOREPTR P_MEM;
  EXTERN STOREPTR S_MEM;
  EXTERN STOREPTR H_MEM;
#else
  EXTERN STOREPTR P;
  EXTERN STOREPTR S;
  EXTERN STOREPTR H;
#endif

EXTERN STOREPTR CP;
EXTERN STOREPTR HB;
EXTERN STOREPTR B0;
EXTERN STOREPTR B;
EXTERN STOREPTR E;
EXTERN STOREPTR TR;
EXTERN RESINX R;
EXTERN RESINX R0;	/* position of first &'ed resource */
EXTERN LONG L;
EXTERN LONG U;
EXTERN CELL RLIST;	/* list of linear resources */
EXTERN CELL R1;		/* resource register */
EXTERN CELL R2;		/* resource register */
EXTERN BOOLEAN T;	/* top flag */

#define REGSIZE 1024
EXTERN CELL a[REGSIZE];		/* a[0] is not used */
#define A(i)        a[i]
#define X(i)        a[i]
#define Y(i)        STORE(E+2+(i))

#define MAXARGS 1024

EXTERN long totalsize;
EXTERN STOREPTR load_base;
EXTERN int halt;
EXTERN int abort_code;
EXTERN int fail;
EXTERN CELL predicate;
EXTERN int num_of_args;
EXTERN int readmode;
EXTERN CELL last_get;
EXTERN jmp_buf am_jmpbuf;
EXTERN long runtime0;
EXTERN long max_usage;

#define	TRACE_CALL	(1)
#define TRACE_AM	(2)
EXTERN int trace;

struct prolog_flag {
    int unknown;	/* error=0, fail=1 */ 
    int occurs_check;   /* off=0, on=1 */
    int max_usage;	/* off=0, on=1 */
} prolog_flag;

/****************************************************************
	LLPAM Macros
****************************************************************/
#define AM_DEREF(x, t, y) \
    {\
        t = TAG(x = (y));\
	while (t == TagREF && x != STORE(REF2PTR(x))) \
	    t = TAG(x = STORE(REF2PTR(x)));\
    }
#define AM_OCCURS_CHECK(x, z) \
	(prolog_flag.occurs_check && am_occurs_check(x, z))
#define AM_ASSIGN(x, z) \
    {\
        STOREPTR p;\
	p = REF2PTR(x);\
        STORE(p) = (z);\
	if (p < HB || (H < p && p < B)) {\
	    STORE(TR++) = x;\
	}\
    }
#define	AM_FAIL \
    {\
        if (B == stack_top) {\
	    halt = YES;\
	    longjmp(am_jmpbuf, 1);\
	}\
        B0 = (STOREPTR)STORE(B+STORE(B)+7);\
        P = (STOREPTR)STORE(B+STORE(B)+4);\
        fail = NO;\
    }
#define AM_STACK_BOT \
	((E>B) ? (E+STORE(E+2)+3) : (B+STORE(B)+CPF_SIZE))

/* #define CPF_SIZE	11 */
#define CPF_SIZE	12
#define	max(x,y)	((x)>(y)?(x):(y))
#define	min(x,y)	((x)<(y)?(x):(y))
#define LISTQ(x)	(TAG(x) == TagLIS)
#define CAR(x)		STORE(LIS2PTR(x))
#define CDR(x)		STORE(LIS2PTR(x)+1)

/****************************************************************
	Subroutines
****************************************************************/
void am_abort(int);
void am_unwind_trail(STOREPTR, STOREPTR);
void am_tidy_trail();
int am_occurs_check(CELL, CELL);
void am_unify(CELL, CELL);
void am_builtin();
void am_init();
void am_free();
int am_interpreter();
int am_load(char *, int);
SYMBOLINX am_make_symbol(char *, int);
void am_define_pred(SYMBOLINX, STOREPTR, int);
void am_exec_indirect(CELL);

/*
 Local Variables:
 mode: c
 c-indent-level: 4
 End:
*/
