/****************************************************************
	LLPAM Emulator
	(resource management)
	by Naoyuki Tamura (tamura@kobe-u.ac.jp)
	   Dept. CS, Kobe University
	Last modified by Mutsunori Banbara on Janly 27 2000
	(banbara@cs32.scitec.kobe-u.ac.jp)
****************************************************************/
#include <stdio.h>

#define	EXTERN extern
#include "llpam.h"
#include "resource.h"

#define AM_INSTR void
/* #define	not_used(k)    (RES(k).level == L) */
#define	not_consumed(i)\
	!(((RES(i).level == U) && (RES(i).deadline == 0)) || (T && (RES(i).level == L) && (RES(i).deadline == L)))
/* checks RES(i) is consumable or not */
#define consumable1(i) \
	((! RES(i).out_of_scope) && ((RES(i).level == 0) || (RES(i).level == L)))
/* checks RES(i) is consumable and has predicate symbol f */
#define consumable(f,i)  ((RES(i).pred == f) && consumable1(i))
#define CON_HASH(x)	(SYMBOL(CON2SYM(x)).hash_value)
#define STR_HASH(x)	(SYMBOL(CON2SYM(STORE(STR2PTR(x)))).hash_value)
#define	ADD_HASH(x, y)	(((x) << 8) + (y))

/****************************************************************
	Subroutines
****************************************************************/
INLINE ULONG am_hash_one_level(CELL x)
{
    /*
       x should be dereferenced
    */
    switch (TAG(x)) {
      case TagREF:
	return(UNDEF);
      case TagINT:
      case TagRES:
	return(x);
      case TagCON:
	return(CON_HASH(x));
      case TagSTR:
	return(STR_HASH(x));
      case TagLIS:
	return(CON_HASH(SYM_DOT));
    }
}

INLINE LONG am_hash(CELL x, ULONG hashsize)
{
    ULONG h;
    CELL arg1;
    int t1;

    /*
       x should be dereferenced
    */
    switch (TAG(x)) {
      case TagREF:
	return(UNDEF);
      case TagINT:
      case TagRES:
      case TagCON:
	return(am_hash_one_level(x) % hashsize);
      case TagSTR:
	arg1 = STORE(STR2PTR(x)+1);
	break;
      case TagLIS:
	arg1 = STORE(LIS2PTR(x));
	break;
    }
    AM_DEREF(arg1, t1, arg1);
    if (t1 == TagREF)
      return(UNDEF);
    h = am_hash_one_level(x);
    h = ADD_HASH(h, am_hash_one_level(arg1));
    return(h % hashsize);
}


/* lookup_hash */
void lookup_hash(SYMBOLINX s)
{
    CELL x;
    int tag;
    ULONG h;

    if (num_of_args == 0) {
	h = SYMBOL(s).hash_value % HASHSIZE;
	R1 = HASH(h);
	R2 = SYM_NIL;
    }
    else {
	AM_DEREF(x, tag, A(1));
	if (tag == TagREF) {
	    R1 = SYMBOL(s).res;
	    R2 = SYM_NIL;
	}
	else {
	    h = SYMBOL(s).hash_value;
	    h = ADD_HASH(h, am_hash_one_level(x)) % HASHSIZE;
	    R1 = HASH(h);
	    R2 = SYMBOL(s).res2;
	}
    }
}


/* updates the level of RES(i) to l */
INLINE void changelevel(RESINX i, LONG l)
{
    RESINX k;

    if (RES(i).level != l) {
#ifdef DEBUG
	if ((trace & TRACE_AM) != 0) {
	    printf("Change Level ");
	    am_print_term(RES(i).head);
	    printf(" : ");
	    printf("%ld", RES(i).level);
	    printf(" -> ");
	    printf("%ld\n", l);
	}
#endif
	STORE(TR++) = MAKETOP(TRUE);
	STORE(TR++) = MAKERES(i);
	STORE(TR++) = MAKEINT(RES(i).level);
	for (k = RES(i).s1; k < RES(i).s2; k++)
	    RES(k).level = l;
    }
}


INLINE void changedeadline(RESINX i, LONG l)
{
    RESINX k;

    if (RES(i).deadline != l) {
#ifdef DEBUG
	if ((trace & TRACE_AM) != 0) {
	    printf("Change Deadline ");
	    am_print_term(RES(i).head);
	    printf(" : ");
	    printf("%ld", RES(i).deadline);
	    printf(" -> ");
	    printf("%ld\n", l);
	}
#endif 
	STORE(TR++) = MAKETOP(FALSE);
	STORE(TR++) = MAKERES(i);
	STORE(TR++) = MAKEINT(RES(i).deadline);
	for (k = RES(i).s1; k < RES(i).s2; k++)
	    RES(k).deadline = l;
    }
}


INLINE void change(LONG l1, LONG l2)
{
    CELL p;
    RESINX i;

    for (p = RLIST; LISTQ(p); p = CDR(p)) {
        i = RES2RESINX(CAR(p));
	if (! RES(i).out_of_scope && RES(i).level == l1) {
	    changelevel(i, l2);
	}
    }
}


INLINE void changepair(LONG l1, LONG d1, LONG l2, LONG d2)
{
    CELL p;
    RESINX i;

    for (p = RLIST; LISTQ(p); p = CDR(p)) {
        i = RES2RESINX(CAR(p));
	if (! RES(i).out_of_scope
	    && RES(i).level == l1 && RES(i).deadline == d1) {
	    changelevel(i, l2);
	    changedeadline(i, d2);
	}
    }
}


INLINE BOOLEAN consumed(LONG l)
{
    CELL p;
    RESINX i;

    for (p = RLIST; LISTQ(p); p = CDR(p)) {
        i = RES2RESINX(CAR(p));
	if (! RES(i).out_of_scope && RES(i).deadline == l)
	    return FALSE;
    }
    return TRUE;
}


/* consumes RES(i) */
INLINE void consume(RESINX i)
{
    if (RES(i).level == 0)
	return;
    changelevel(i, U);
    changedeadline(i, 0);
}


INLINE SYMBOLINX am_register_resource(CELL x)
{
    SYMBOLINX pred;
    int t;
    LONG h;

    switch (TAG(x)) {
      case TagCON:
	pred = CON2SYM(x);
	break;
      case TagSTR:
	pred = CON2SYM(STORE(STR2PTR(x)));
	break;
      default:
	am_abort(1012);
	return;
    }
    /* add to res */
    if (SYMBOL(pred).res == UNDEF)
	SYMBOL(pred).res = SYM_NIL;
    STORE(H) = MAKERES(R);
    STORE(H+1) = SYMBOL(pred).res;
    SYMBOL(pred).res = MAKELIS(H);
    H += 2;
    STORE(TR++) = MAKELIS((STOREPTR)&SYMBOL(pred).res);
    /* add to HASH or res2 */
    if (SYMBOL(pred).res2 == UNDEF)
      SYMBOL(pred).res2 = SYM_NIL;
    if ((h = am_hash(x, HASHSIZE)) == UNDEF) {
	STORE(H) = MAKERES(R);
	STORE(H+1) = SYMBOL(pred).res2;
	SYMBOL(pred).res2 = MAKELIS(H);
	H += 2;
	STORE(TR++) = MAKELIS((STOREPTR)&SYMBOL(pred).res2);
    }
    else {
	/* if (HASH(h) == UNDEF)
	  HASH(h) = SYM_NIL; */
#ifdef DEBUG
	if (TAG(HASH(h)) == TagLIS) {
	    printf("Hash(");
	    am_print_term(x);
	    printf(") = %ld is duplicating with Hash(", h);
	    am_print_term(RES(INT2LONG(CAR(HASH(h)))).head);
	    printf(")\n");
	}
#endif
	STORE(H) = MAKERES(R);
	STORE(H+1) = HASH(h);
	HASH(h) = MAKELIS(H);
	H += 2;
	STORE(TR++) = MAKELIS((STOREPTR)&HASH(h));
    }
#ifdef DEBUG
    if ((trace & TRACE_AM) != 0) {
	printf("RES(%ld)=", R);
	am_print_term(x);
	printf(", SYMBOL(");
	am_print_functor_arity(MAKECON(pred));
	printf(").res=");
	am_print_term(SYMBOL(pred).res);
	printf(", SYMBOL(");
	am_print_functor_arity(MAKECON(pred));
	printf(").res2=");
	am_print_term(SYMBOL(pred).res2);
	if (h != UNDEF) {
	    printf(", HASH(%ld)=", h);
	    am_print_term(HASH(h));
	}
	printf("\n");
    }
#endif
    return pred;
}


INLINE SYMBOLINX am_register_resource_term(CELL x)
{
    SYMBOLINX pred;
    int t;
    LONG h;

    switch (TAG(x)) {
      case TagCON:
	pred = CON2SYM(x);
	break;
      case TagSTR:
	pred = CON2SYM(STORE(STR2PTR(x)));
	break;
      default:
	AM_FAIL;
	return;
    }
    /* add to res3 */
    if (SYMBOL(pred).res3 == UNDEF)
	SYMBOL(pred).res3 = SYM_NIL;
    STORE(H) = MAKERES(R);
    STORE(H+1) = SYMBOL(pred).res3;
    SYMBOL(pred).res3 = MAKELIS(H);
    H += 2;
    STORE(TR++) = MAKELIS((STOREPTR)&SYMBOL(pred).res3);
    /* add to HASH or res4 */
    if (SYMBOL(pred).res4 == UNDEF)
	SYMBOL(pred).res4 = SYM_NIL;
    if ((h = am_hash(x, HASHSIZE)) == UNDEF) {
	STORE(H) = MAKERES(R);
	STORE(H+1) = SYMBOL(pred).res4;
	SYMBOL(pred).res4 = MAKELIS(H);
	H += 2;
	STORE(TR++) = MAKELIS((STOREPTR)&SYMBOL(pred).res4);
    }
    else {
	/* if (HASH(h) == UNDEF)
	  HASH(h) = SYM_NIL; */
#ifdef DEBUG
	if (TAG(HASH2(h)) == TagLIS) {
	    printf("Hash2(");
	    am_print_term(x);
	    printf(") = %ld is duplicating with Hash2(", h);
	    am_print_term(RES(INT2LONG(CAR(HASH2(h)))).head);
	    printf(")\n");
	}
#endif
	STORE(H) = MAKERES(R);
	STORE(H+1) = HASH2(h);
	HASH2(h) = MAKELIS(H);
	H += 2;
	STORE(TR++) = MAKELIS((STOREPTR)&HASH2(h));
    }
#ifdef DEBUG
    if ((trace & TRACE_AM) != 0) {
	printf("RES(%ld)=", R);
	am_print_term(x);
	printf(", SYMBOL(");
	am_print_functor_arity(MAKECON(pred));
	printf(").res3=");
	am_print_term(SYMBOL(pred).res3);
	printf(", SYMBOL(");
	am_print_functor_arity(MAKECON(pred));
	printf(").res4=");
	am_print_term(SYMBOL(pred).res4);
	if (h != UNDEF) {
	    printf(", HASH2(%ld)=", h);
	    am_print_term(HASH2(h));
	}
	printf("\n");
    }
#endif
    return pred;
}



INLINE void am_end_imp_sub(CELL x1, CELL x2)
{
    int t1, t2;
    RESINX from, to;
    RESINX k;

    if (TAG(x1) != TagRES || TAG(x2) != TagRES) {
	fail = YES;
	return;
    }
    from = RES2RESINX(x1);
    to = RES2RESINX(x2);
    if (from >= to)
      return;
    for (k = from; k < to; k++) {
	if (not_consumed(k)) {
	    fail = YES;
	    return;
	}
    }
    for (k = from; k < to; k++) {
	RES(k).out_of_scope = YES;
    }
    STORE(TR++) = MAKERES(from);
    STORE(TR++) = MAKERES(to);
}


INLINE void am_end_exp_imp_sub(CELL x1, CELL x2)
{
    int t1, t2;
    RESINX from, to;
    RESINX k;

    if (TAG(x1) != TagRES || TAG(x2) != TagRES) {
	fail = YES;
	return;
    }
    from = RES2RESINX(x1);
    to = RES2RESINX(x2);
    if (from >= to)
      return;
    for (k = from; k < to; k++) {
	RES(k).out_of_scope = YES;
    }
    STORE(TR++) = MAKERES(from);
    STORE(TR++) = MAKERES(to);
}


INLINE void am_exec_resource_indirect(CELL goal)
{
    STOREPTR p;
    SYMBOLINX pred;
    CELL y, arg1;
    int tag;
    ULONG h;

    switch (TAG(goal)) {
    case TagCON:
	num_of_args = 0;
	predicate = goal;
	break;
    case TagSTR:
	p = STR2PTR(goal);
	predicate = STORE(p);
	num_of_args = SYMBOL(CON2SYM(predicate)).arity;
	arg1 =  STORE(p+1);
	break;
    default:
	printf("Indirect call resource error\n");
	am_abort(2);
	break;
    }
    pred = CON2SYM(predicate);
    B0 = B;
    if (TAG(SYMBOL(pred).res3) != TagLIS) {
	AM_FAIL;
	return;
    } else {
	if (num_of_args == 0) {
	    h = SYMBOL(pred).hash_value % HASHSIZE;
	    R1 = HASH2(h);
	    R2 = SYM_NIL;
	}
	else {
	    AM_DEREF(y, tag, arg1);
	    if (tag == TagREF) {
		R1 = SYMBOL(pred).res3;
		R2 = SYM_NIL;
	    }
	    else {
		h = SYMBOL(pred).hash_value;
		h = ADD_HASH(h, am_hash_one_level(y)) % HASHSIZE;
		R1 = HASH2(h);
		R2 = SYMBOL(pred).res4;
	    }
	}
    }
    A(1) = goal;
    A(2) = predicate;
    /* SYM_RES is a builtin predicate (refer built.llp) */
    predicate = SYM_RES;
    num_of_args = 2;
    P = SYMBOL(CON2SYM(predicate)).codeaddr;
    /* Perhaps not needed */
    if (P == PRED_UNDEF) {
	printf("%s/%d is undefined\n",
	       SYMBOL(CON2SYM(predicate)).print_name, num_of_args);
	am_abort(10);
    }	
}


/****************************************************************
	Instructions
****************************************************************/

/* debug_dmp_r */
AM_INSTR am_debug_dmp_r()
{
    RESINX i;
    res_struct x;

    printf("R=%ld L=%ld, U=%ld, T=%d\n", R, L, U, T);
    for (i = 0; i < R; i++) {
	x = RES(i);
	printf("  RES[%3ld]\n", (LONG)i);
	printf("      pred  = %s/", SYMBOL(x.pred).print_name);
	printf("%d\n", SYMBOL(x.pred).arity);
	printf("      s1    = %d\n", x.s1);
	printf("      s2    = %d\n", x.s2);
	printf("      level = %d\n", x.level);
	printf("      deadline = %d\n", x.deadline);
	printf("      out_of_scope = %d\n", x.out_of_scope);
	printf("      head  = "); am_print_term(x.head); printf("\n");
	printf("      body  = "); am_print_term(x.body); printf("\n");
	printf("      closure = "); am_print_term(x.closure); printf("\n");
	printf("\n");
    }
    P++;
}


/* debug_dmp_r_reg */
AM_INSTR am_debug_dmp_r_reg()
{
    printf("R1 = "); am_print_term(R1); printf("\n");
    printf("R2 = "); am_print_term(R2); printf("\n");
    P++;
}


/* begin_imp Ai */
AM_INSTR am_begin_imp_a()
{
    A(CARG1) = MAKERES(R);
    R0 = R;
    P++;
}


/* begin_imp Yi */
AM_INSTR am_begin_imp_y()
{
    Y(CARG1) = MAKERES(R);
    R0 = R;
    P++;
}


/* begin_exp_imp Ai */
AM_INSTR am_begin_exp_imp_a()
{
    A(CARG1) = MAKERES(R);
    P++;
}


/* begin_exp_imp Yi */
AM_INSTR am_begin_exp_imp_y()
{
    Y(CARG1) = MAKERES(R);
    P++;
}


/* add_res Ai, Aj */
AM_INSTR am_add_res()
{
    /*
      A(CARG1), A(CARG2) should be dereferenced
    */
    RES(R).level = L;
    RES(R).deadline = L;
    RES(R).out_of_scope = NO;
    RES(R).head = A(CARG1);
    RES(R).body = UNDEF;
    RES(R).closure = A(CARG2);
    RES(R).pred    = am_register_resource(RES(R).head);
    R++;
    P++;
}


/* add_res_term Ai, Aj */
AM_INSTR am_add_res_term()
{
    /*
      A(CARG1), A(CARG2) should be dereferenced
    */
    RES(R).level = L;
    RES(R).deadline = L;
    RES(R).out_of_scope = NO;
    RES(R).head = A(CARG1);
    RES(R).body = A(CARG2);
    RES(R).closure = UNDEF;
    RES(R).pred    = am_register_resource_term(RES(R).head);
    R++;
    P++;
}


/* add_exp_res Ai, Aj */
AM_INSTR am_add_exp_res()
{
    /*
       A(CARG1), A(CARG2) should be dereferenced
    */
    RES(R).level = 0;
    RES(R).deadline = 0;
    RES(R).out_of_scope = NO;
    RES(R).head = A(CARG1);
    RES(R).body = UNDEF;
    RES(R).closure = A(CARG2);
    RES(R).pred    = am_register_resource(RES(R).head);
    R++;
    P++;
}


/* add_exp_res_term Ai, Aj */
AM_INSTR am_add_exp_res_term()
{
    /*
       A(CARG1), A(CARG2) should be dereferenced
    */
    RES(R).level = 0;
    RES(R).deadline = 0;
    RES(R).out_of_scope = NO;
    RES(R).head = A(CARG1);
    RES(R).body = A(CARG2);
    RES(R).closure = UNDEF;
    RES(R).pred    = am_register_resource_term(RES(R).head);
    R++;
    P++;
}


/* mid_imp Ai Aj */
AM_INSTR am_mid_imp_a()
{
    RESINX i;
    
    A(CARG1) = MAKERES(R);
    A(CARG2) = MAKETOP(T);
    T = FALSE;
    for(i = R0; i < R; i++){
	RES(i).s1 = R0;
	RES(i).s2 = R;
    }
    STORE(H) = MAKERES(R0);
    STORE(H+1) = RLIST;
    RLIST = MAKELIS(H);
    H += 2;
    STORE(TR++) = SYM_NIL;
    P++;
}

/* mid_imp Yi Yj */
AM_INSTR am_mid_imp_y()
{
    RESINX i;

    Y(CARG1) = MAKERES(R);
    Y(CARG2) = MAKETOP(T);
    T = FALSE;
    for(i = R0; i < R; i++){
	RES(i).s1 = R0;
	RES(i).s2 = R;
    }
    STORE(H) = MAKERES(R0);
    STORE(H+1) = RLIST;
    RLIST = MAKELIS(H);
    H += 2;
    STORE(TR++) = SYM_NIL;
    P++;
}

/* mid_exp_imp Ai, Aj */
AM_INSTR am_mid_exp_imp_a()
{
    A(CARG1) = MAKERES(R);
    A(CARG2) = MAKETOP(T);
    T = FALSE;
    P++;
}


/* mid_exp_imp Yi, Yj */
AM_INSTR am_mid_exp_imp_y()
{
    Y(CARG1) = MAKERES(R);
    Y(CARG2) = MAKETOP(T);
    T = FALSE;
    P++;
} 


/* more_imp */
AM_INSTR am_more_imp()
{
    RESINX i;
    
    for(i = R0; i < R; i++){
	RES(i).s1 = R0;
	RES(i).s2 = R;
    }
    STORE(H) = MAKERES(R0);
    STORE(H+1) = RLIST;
    RLIST = MAKELIS(H);
    H += 2;
    STORE(TR++) = SYM_NIL;
    R0 = R;
    P++;
}


/* end_imp Ai, Aj, Ak */
AM_INSTR am_end_imp_a()
{
    /*
       A(CARG1), A(CARG2), A(CARG3) should be dereferenced
    */
    am_end_imp_sub(A(CARG1), A(CARG2));
    if (fail) {
	AM_FAIL;
	return;
    }
    if (TAG(A(CARG3)) != TagTOP) {
	printf("Error: am_end_imp_a\n");
	am_abort(1011);
	return;
    }
    T = (TOP2BOOL(A(CARG3)) || T);
    P += 2;
}


/* end_imp Yi,Yj,Yk */
AM_INSTR am_end_imp_y()
{
    /*
       Y(CARG1), Y(CARG2), Y(CARG3) should be dereferenced
    */
    am_end_imp_sub(Y(CARG1), Y(CARG2));
    if (fail) {
	AM_FAIL;
	return;
    }
    if (TAG(Y(CARG3)) != TagTOP) {
	printf("Error: am_end_imp_y\n");
	am_abort(1011);
	return;
    }
    T = (TOP2BOOL(Y(CARG3)) || T);
    P += 2;
}


/* end_exp_imp Ai,Aj,Ak */
AM_INSTR am_end_exp_imp_a()
{
    /*
       A(CARG1), A(CARG2), A(CARG3) should be dereferenced
    */
    am_end_exp_imp_sub(A(CARG1), A(CARG2));
    if (fail) {
	AM_FAIL;
	return;
    }
    if (TAG(A(CARG3)) != TagTOP) {
	printf("Error: am_end_imp_a\n");
	am_abort(1011);
	return;
    }
    T = (TOP2BOOL(A(CARG3)) || T);
    P += 2;
}


/* end_exp_imp Yi,Yj,Yk */
AM_INSTR am_end_exp_imp_y()
{
    /*
       Y(CARG1), Y(CARG2), Y(CARG3) should be dereferenced
    */
    am_end_exp_imp_sub(Y(CARG1), Y(CARG2));
    if (fail) {
	AM_FAIL;
	return;
    }
    if (TAG(Y(CARG3)) != TagTOP) {
	printf("Error: am_end_imp_y\n");
	am_abort(1011);
	return;
    }
    T = (TOP2BOOL(Y(CARG3)) || T);
    P += 2;
}


/* begin_with Ai */
AM_INSTR am_begin_with_a()
{
    U--;
    A(CARG1) = MAKETOP(T);
    T = FALSE;
    P++;
}

/* begin_with Yi */
AM_INSTR am_begin_with_y()
{
    U--;
    Y(CARG1) = MAKETOP(T);
    T = FALSE;
    P++;
}


/* mid_with Ai */
AM_INSTR am_mid_with_a()
{
    changepair(U,0,L+1,L+1);
    if (T)
	change(L,L+1);
    L++;
    U++;
    A(CARG1) = MAKETOP(T);
    T = FALSE;
    P++;
}


/* mid_with Yi */
AM_INSTR am_mid_with_y()
{
    changepair(U,0,L+1,L+1);
    if (T)
	change(L,L+1);
    L++;
    U++;
    Y(CARG1) = MAKETOP(T);
    T = FALSE;
    P++;
}


/* end_with Ai,Aj */
AM_INSTR am_end_with_a()
{
    BOOLEAN t1,t2; 
    if (TAG(A(CARG1)) != TagTOP || TAG(A(CARG2)) != TagTOP) {
	printf("Error: am_end_with_a\n");
	am_abort(1011);
	return;
    }
    t1 = TOP2BOOL(A(CARG1));
    t2 = TOP2BOOL(A(CARG2));
    L--;
    if (T)
	changepair(L+1,L+1,U,0);
    else if (! consumed(L+1)) {
	AM_FAIL;
	return;
    }
    if (t2)
	change(L+1,L);
    T = (t1 || (t2 && T));
    P++;
}


/* end_with Yi,Yj */
AM_INSTR am_end_with_y()
{
    BOOLEAN t1,t2; 
    if (TAG(Y(CARG1)) != TagTOP || TAG(Y(CARG2)) != TagTOP) {
	printf("Error: am_end_with_y\n");
	am_abort(1011);
	return;
    }
    t1 = TOP2BOOL(Y(CARG1));
    t2 = TOP2BOOL(Y(CARG2));
    L--;
    if (T)
	changepair(L+1,L+1,U,0);
    else if (! consumed(L+1)) {
	AM_FAIL;
	return;
    }
    if (t2)
	change(L+1,L);
    T = (t1 || (t2 && T));
    P++;
}


/* begin_bang Ai */
AM_INSTR am_begin_bang_a()
{
    L++;
    A(CARG1) = MAKETOP(T);
    P++;
}


/* begin_bang Yi */
AM_INSTR am_begin_bang_y()
{
    L++;
    Y(CARG1) = MAKETOP(T);
    P++;
}


/* end_bang Ai */
AM_INSTR am_end_bang_a()
{
    if (TAG(A(CARG1)) != TagTOP) {
	printf("Error: am_end_bang_a\n");
	am_abort(1011);
	return;
    }
    L--;
    T = TOP2BOOL(A(CARG1));
    P++;
}


/* end_bang Yi */
AM_INSTR am_end_bang_y()
{
    if (TAG(Y(CARG1)) != TagTOP) {
	printf("Error: am_end_bang_y\n");
	am_abort(1011);
	return;
    }
    L--;
    T = TOP2BOOL(Y(CARG1));
    P++;
}


/* unify_resource Ai, Aj */
AM_INSTR am_unify_resource()
{
    RESINX i;

    i = RES2RESINX(A(CARG1));
    am_unify(RES(i).head, A(CARG2));
    if (fail) {
	AM_FAIL;
	return;
    }
    consume(i);
    P++;
}


/* resource_body Ai,Aj */
AM_INSTR am_resource_body()
{
    RESINX i;

    i = RES2RESINX(A(CARG1));
    A(CARG2) = RES(i).body;
    P++;
}


/* top */
AM_INSTR am_top()
{
    T = TRUE;
    P++;
}


/* erase */
/*
AM_INSTR am_erase()
{
    CELL p;
    RESINX i;

    for (p = RLIST; LISTQ(p); p = CDR(p)) {
        i = RES2RESINX(CAR(p));
	if (! RES(i).out_of_scope && RES(i).level == L) {
	    changelevel(i, U);
	    changedeadline(i, 0);
	}
    }
    P++;
}
*/
AM_INSTR am_erase()
{
    T = TRUE;
    P++;
}


/* put_closure L, n, Ai */
AM_INSTR am_put_closure()
{
    STOREPTR label;

    label = INX2PTR(CARGW(1));
    A(CARG1)   = MAKECLO(H);
    STORE(H++) = (CELL)label;
    STORE(H++) = CARGW(2);
    readmode = NO;
    P += 3;
}


/* execute_closure Aj */
AM_INSTR am_execute_closure()
{
    CELL x;

    x = A(CARG1);
    if ( TAG(x) != TagCLO ){
	printf("Error: am_execute_closure\n");
	am_abort(1011);
    }
    B0 = B;
    readmode = YES;
    S = CLO2PTR(x);
    P = (STOREPTR)STORE(S);
    S += 2;
}


/* consume Ai Aj */
AM_INSTR am_consume()
{
    CELL x;

    x = A(CARG1);
    if ( TAG(x) != TagRES ){
	printf("Error: am_consume\n");
	am_abort(1011);
    }
    consume(RES2INX(x));
    A(CARG2) = RES(RES2INX(x)).closure; 
    P++;
}


/* pickup_resource f/n, Ai, L */
AM_INSTR am_pickup_resource()
{
    SYMBOLINX s;
    STOREPTR label;

    s = CON2SYM(CARGW(1));
    label = INX2PTR(CARGW(2));
    for (; LISTQ(R1); R1 = CDR(R1)) {
	if ( consumable(s, RES2RESINX(CAR(R1))) ){ 
	    A(CARG1) = CAR(R1);
	    R1 = CDR(R1);
	    P += 3;
	    return;
	}
    }
    for (; LISTQ(R2); R2 = CDR(R2)) {
	if ( consumable(s, RES2RESINX(CAR(R2))) ){
	    A(CARG1) = CAR(R2);
	    R2 = CDR(R2);
	    P += 3;
	    return;
	}
    }
    P = label;
}


/* pickup_resource_term Ai, Aj, L */
AM_INSTR am_pickup_resource_term()
{
    SYMBOLINX s;
    STOREPTR label;

    s = CON2SYM(A(CARG1));
    label = INX2PTR(CARGW(1));
    for (; LISTQ(R1); R1 = CDR(R1)) {
	if ( consumable(s, RES2RESINX(CAR(R1))) ){ 
	    A(CARG2) = CAR(R1);
	    R1 = CDR(R1);
	    P += 2;
	    return;
	}
    }
    for (; LISTQ(R2); R2 = CDR(R2)) {
	if ( consumable(s, RES2RESINX(CAR(R2))) ){
	    A(CARG2) = CAR(R2);
	    R2 = CDR(R2);
	    P += 2;
	    return;
	}
    }
    P = label;
}


/* if_no_resource L */
AM_INSTR am_if_no_resource()
{
    STOREPTR label;

    label = INX2PTR(CARGW(1));
    for (; LISTQ(R1); R1 = CDR(R1)) {
	if (consumable1(RES2RESINX(CAR(R1))))
	  break;
    }
    for (; LISTQ(R2); R2 = CDR(R2)) {
	if (consumable1(RES2RESINX(CAR(R2))))
	  break;
    }
    if (R1 == SYM_NIL) {
	R1 = R2;
	R2 = SYM_NIL;
    }
    if (TAG(R1) != TagLIS) {
	P = label;
	return;
    }	
    P += 2;
}


/* try_resource L */
AM_INSTR am_try_resource()
{
    STOREPTR label, newB;
    int i, n;

    label = INX2PTR(CARGW(1));
    newB = AM_STACK_BOT;
    n = STORE(newB) = num_of_args + 2;
    for (i = 1; i <= num_of_args; i++)
	STORE(newB+i) = A(i);
    STORE(newB+n-1) = R1;
    STORE(newB+n) = R2;
    STORE(newB+n+1) = (CELL)E;
    STORE(newB+n+2) = (CELL)CP;
    STORE(newB+n+3) = (CELL)B;
    STORE(newB+n+4) = (CELL)(P + 2);
    STORE(newB+n+5) = (CELL)TR;
    STORE(newB+n+6) = (CELL)H;
    STORE(newB+n+7) = (CELL)B0;
    STORE(newB+n+8) = R;
    STORE(newB+n+9) = L;
    STORE(newB+n+10) = U;
    STORE(newB+n+11) = (CELL)T; /* added */
    B = newB;
    HB = H;
    P = label;
}


/* restore_resource */
AM_INSTR am_restore_resource()
{
    int i, n;

    n = STORE(B);
    num_of_args = n - 2;
    for (i = 1; i <= num_of_args; i++)
      A(i) = STORE(B+i);
    R1 = STORE(B+n-1);
    R2 = STORE(B+n);
    E = (STOREPTR)STORE(B+n+1);
    CP = (STOREPTR)STORE(B+n+2);
    am_unwind_trail((STOREPTR)STORE(B+n+5), TR);
    TR = (STOREPTR)STORE(B+n+5);
    H = (STOREPTR)STORE(B+n+6);
    R = STORE(B+n+8);
    L = STORE(B+n+9);
    U = STORE(B+n+10);
    T = (BOOLEAN)STORE(B+n+11);
    P++;
}


/* retry_resource_else L */
AM_INSTR am_retry_resource_else()
{
    STOREPTR label;
    int n;

    label = INX2PTR(CARGW(1));
    n = STORE(B);
    STORE(B+n-1) = R1;
    STORE(B+n) = R2;    
    STORE(B+n+4) = (CELL)label;
    HB = H;
    P+=2;
}


/* trust_resource L */
AM_INSTR am_trust_resource()
{
    STOREPTR label;
    int n;

    label = INX2PTR(CARGW(1));
    n  = STORE(B);
    B  = (STOREPTR)STORE(B+n+3);
    HB = (STOREPTR)STORE(B+STORE(B)+6);
    P  = label;
}


/* execute_resource_indirect Ai */
AM_INSTR am_execute_resource_indirect()
{
    SYMBOLINX pred;
    /*
       A(CARG1) should be dereferenced beforehand
    */
    am_exec_resource_indirect(A(CARG1));
}






/*
 Local Variables:
 mode: c
 c-indent-level: 4
 End:
*/
