/**
 * @(#)Pentomino.java
 * @author Naoyuki Tamura (tamura@kobe-u.ac.jp)
 * Dept. CS, Kobe University
 *
 * Pentomino Puzzle
 */
import java.awt.*;
import java.io.*;
import java.util.*;

public class Pentomino extends PrologFrame {
    private PentominoPanel pentominoPanel;
    private TextField M_Field;

    public Pentomino() {
	setSize(500, 400);
	setLayout(new BorderLayout());
	add("North", new Label("Pentominos", Label.CENTER));
	pentominoPanel = new PentominoPanel();
	add("Center", pentominoPanel);
	Panel p = new Panel();
	add("South", p);
	p.add(new Label("Rows="));
	M_Field = new TextField("6", 2);
	p.add(M_Field);
	control = new PrologControlPanel("llp -ejava pentomino");
	p.add(control);
	validate();
    }

    void sendArgument() {
	int M;
	try {
	    M = Integer.parseInt(M_Field.getText());
	} catch (NumberFormatException e) {
	    M = 6;
	    M_Field.setText(Integer.toString(M));
	}
	control.sendArgument(Integer.toString(M));
    }

    void receiveResult() {
	String result = control.receiveResult();
	if (result == null || result.equals("DONE")) {
	    return;
	}
	StringTokenizer st = new StringTokenizer(result, ",");
	try {
	    int M = Integer.parseInt(st.nextToken());
	    int N = Integer.parseInt(st.nextToken());
	    char pentomino[][] = new char[M][N];
	    for (int i = 0; i < M; i++) {
		for (int j = 0; j < N; j++) {
		    pentomino[i][j] = st.nextToken().charAt(0);
		}
	    }
	    pentominoPanel.setPentomino(M, N, pentomino);
	} catch (NumberFormatException e) {
	} catch (StringIndexOutOfBoundsException e) {}
    }

    public static void main(String args[]) {
	new Pentomino();
    }
}


class PentominoPanel extends Panel {
    private int M;
    private int N;
    private Color pentomino[][] = null;

    public void setPentomino(int m, int n, char p[][]) {
	M = m;
	N = n;
	pentomino = new Color[M][N];
	for (int i = 0; i < M; i++) {
	    for (int j = 0; j < N; j++) {
		Color c = Color.white;
		switch (p[i][j]) {
		  case 'F': c = Color.blue; break;
		  case 'I': c = Color.green; break;
		  case 'L': c = Color.red; break;
		  case 'N': c = Color.cyan; break;
		  case 'P': c = Color.magenta; break;
		  case 'T': c = Color.yellow; break;
		  case 'U': c = Color.orange; break;
		  case 'V': c = Color.pink; break;
		  case 'W': c = Color.black; break;
		  case 'X': c = Color.gray; break;
		  case 'Y': c = new Color(64,128,128); break;
		  case 'Z': c = new Color(128,64,128); break;
		}
		pentomino[i][j] = c;
	    }
	}
	repaint();
    }

    public void paint(Graphics g) {
	if (pentomino == null)
	  return;
	int w = getSize().width;
	int h = getSize().height;
	w = w - 10;
	h = h - 10;
	int cellSize = Math.min(w/N, h/M);
	int x0 = (w - N * cellSize) / 2 + 5;
	int y0 = (h - M * cellSize) / 2 + 5;
	for (int i = 0; i < M; i++)
	  for (int j = 0; j < N; j++) {
	      g.setColor(pentomino[i][j]);
	      g.fillRect(x0+j*cellSize, y0+i*cellSize, cellSize, cellSize);
	  }
    }

}

