/**
 * @(#)Queens.java
 * @author Naoyuki Tamura (tamura@kobe-u.ac.jp)
 * Dept. CS, Kobe University
 *
 * N-Queens
 */
import java.awt.*;
import java.io.*;
import java.util.*;

public class Queens extends PrologFrame {
    private QueensPanel queensPanel;
    private TextField N_Field;
    private int N = 8;

    public Queens() {
	setSize(500, 400);
	setLayout(new BorderLayout());
	add("North", new Label("N-Queens", Label.CENTER));
	queensPanel = new QueensPanel(N);
	add("Center", queensPanel);
	Panel p = new Panel();
	add("South", p);
	p.add(new Label("N="));
	N_Field = new TextField(Integer.toString(N), 2);
	p.add(N_Field);
	control = new PrologControlPanel("llp -ejava queens");
	p.add(control);
	validate();
    }

    void sendArgument() {
	try {
	    N = Integer.parseInt(N_Field.getText());
	} catch (NumberFormatException e) {
	    N = 1;
	    N_Field.setText(Integer.toString(N));
	}
	queensPanel.setN(N);
	control.sendArgument(Integer.toString(N));
    }

    void receiveResult() {
	String result = control.receiveResult();
	if (result == null || result.equals("DONE")) {
	    return;
	}
	StringTokenizer st = new StringTokenizer(result, ",");
	int queens[] = new int[N];
	try {
	    for (int i = 0; i < N; i++)
	      queens[i] = Integer.parseInt(st.nextToken()) - 1;
	} catch (NumberFormatException e) {}
	queensPanel.setQueens(queens);
    }

    public static void main(String args[]) {
	new Queens();
    }
}


class QueensPanel extends Panel {
    private int N = 0;
    private int queens[] = null;
    private int cellSize = 0;
    private Image boardImage = null;
    private int boardN = 0;

    public QueensPanel(int n) {
	setN(n);
    }

    public void setN(int n) {
	if (n > 0) {
	    N = n;
	    setQueens(null);
	    repaint();
	}
    }

    public void setQueens(int q[]) {
	queens = q;
	repaint();
    }

    private void drawBoard(Graphics g, int x0, int y0) {
	int size = N * cellSize;
	if (boardImage == null || N != boardN
	    || boardImage.getWidth(null) != size
	    || boardImage.getHeight(null) != size) {
	    boardImage = createImage(size, size);
	    Graphics b = boardImage.getGraphics();
	    b.setColor(Color.white);
	    b.fillRect(0, 0, size, size);
	    b.setColor(Color.gray);
	    for (int i = 0; i < N; i++)
	      for (int j = 0; j < N; j++)
		if ((i + j) % 2 == 1) {
		    b.fillRect(i*cellSize, j*cellSize, cellSize, cellSize);
		}
	    boardN = N;
	}
	g.drawImage(boardImage, x0, y0, null);
    }
	    
    public void paint(Graphics g) {
	if (N < 1)
	  return;
	int w = getSize().width;
	int h = getSize().height;
	cellSize = Math.min(w, h) / N;
	int x0 = (w - N * cellSize) / 2;
	int y0 = (h - N * cellSize) / 2;
	drawBoard(g, x0, y0);
	if (queens == null || N != queens.length)
	  return;
	int d = (8*cellSize)/10;
	for (int i = 0; i < N; i++) {
	    int j = queens[i];
	    int x = x0 + i * cellSize + (cellSize - d) / 2;
	    int y = y0 + j * cellSize + (cellSize - d) / 2;
	    g.setColor(Color.orange);
	    g.fillArc(x, y, d, d, 0, 360);
	    g.setColor(Color.black);
	    g.drawArc(x, y, d, d, 0, 360);
	}
    }
}


