/**
 * @(#)Triangle.java
 * @author Naoyuki Tamura (tamura@kobe-u.ac.jp)
 * Dept. CS, Kobe University
 *
 * Triangle applet
 */
import java.awt.*;
import java.io.*;
import java.util.*;

public class Triangle extends PrologFrame {
    private TrianglePanel trianglePanel;
    private TextField N_Field;
    private int N = 4;

    public Triangle() {
	setSize(500, 400);
	setLayout(new BorderLayout());
	add("North", new Label("Triangle", Label.CENTER));
	trianglePanel = new TrianglePanel(N);
	add("Center", trianglePanel);
	Panel p = new Panel();
	add("South", p);
	p.add(new Label("N="));
	N_Field = new TextField(Integer.toString(N), 2);
	p.add(N_Field);
	control = new PrologControlPanel("llp -ejava triangle");
	p.add(control);
	validate();
    }

    void sendArgument() {
	try {
	    N = Integer.parseInt(N_Field.getText());
	} catch (NumberFormatException e) {
	    N = 1;
	    N_Field.setText(Integer.toString(N));
	}
	trianglePanel.setN(N);
	control.sendArgument(Integer.toString(N));
    }

    void receiveResult() {
	String result = control.receiveResult();
	if (result == null || result.equals("DONE")) {
	    return;
	}
	StringTokenizer st = new StringTokenizer(result, ",");
	int M = N*(N+1)/2;
	int triangle[] = new int[M];
	try {
	    for (int i = 0; i < M; i++)
	      triangle[i] = Integer.parseInt(st.nextToken());
	} catch (NumberFormatException e) {}
	trianglePanel.setTriangle(triangle);
    }

    public static void main(String args[]) {
	new Triangle();
    }
}


class TrianglePanel extends Panel {
    private int N = 0;
    private int triangle[] = null;

    public TrianglePanel(int n) {
	setN(n);
    }

    public void setN(int n) {
	if (n > 0) {
	    N = n;
	    setTriangle(null);
	    repaint();
	}
    }

    public void setTriangle(int t[]) {
	triangle = t;
	repaint();
    }

    public void paint(Graphics g) {
	if (N < 1 || triangle == null)
	  return;
	int w = getSize().width;
	int h = getSize().height;
	int r = Math.min(w, h)/(2*N);
	int x0 = (w - 2*r*N) / 2;
	int y0 = (h - 2*r*N) / 2;
	Font font = g.getFont();
	font = new Font(font.getName(), font.getStyle(), 24);
	g.setFont(font);
	FontMetrics fontMetrics = g.getFontMetrics();
	int k = 0;
	for (int j = 0; j < N; j++) {
	    for (int i = 0; i < N-j; i++) {
		int x = x0 + 2*r*i + r*j;
		int y = y0 + (int)(2*r*(N-1-j)*1.732/2);
		g.setColor(Color.white);
		g.fillArc(x, y, 2*r, 2*r, 0, 360);
		String str = Integer.toString(triangle[k]);
		x = x + r - fontMetrics.stringWidth(str) / 2;
		y = y + r + fontMetrics.getAscent() / 2;
		g.setColor(Color.black);
		g.drawString(str, x, y);
		k++;
	    }
	}
    }
}


